#include "defs.h"

/*
 *	XENIX 8086 Debugger	Instruction disasembly
 *
 * Originaly written by hans, Aug/Sept, 1981.
 *
 * Modification History:
 *	arw	30 Dec 1981	M000
 *              - Modified ACMEM instruction type to always print
 *		  two bytes of memory address symbolically and use
 *		  the appropriate register name.
 *		- Modified ACIMM to print the appropriate register name.
 *		- Fixed hex() to accept an integer argument.
 *		- Changed various routines to use integer parameters to
 *		  avoid sign extension problems.
 *	arw	 5 Jan 1982	M001
 *		- Corrected table entry for `ff 28' instruction.
 *		- Corrected table entry for `8f' instruction.
 *	arw	19 Jan 1982	M002
 *		- Put in kludge to print string names correctly.
 *		- Fixed a problem with direction bit in MRNORM.
 *	arw	20 Apr 1982	M003
 *		- Fixed hex() to properly indicate immediate and
 *		  short immediate values.  Also fixed to use syntax
 *		  similar to that used by the assembler.
 *	arw	 9 Jun 1982	M004
 *		- Fixed disassembly of intersegment call, jump and return,
 *		  nop and segment load and store.
 *	Hans	19 Dec 1982	M005
 *		reordered segment addresses for DISP4
 *		and made it do a psymoff
 *		removed data1, etc, which were vestages of an older approach
 *	Hans	24 Feb, 83	M006
 *		fixed lea, les, and lds once and for all by removing
 *		them from being a special case of MRNORM and
 *		giving them their own class.
 *	Hans	2 Mar, 83	M007
 *		some segmentation support
 *	Hans	9 Mar 83	M008
 *		removed some wrong hackery.  I dont know what was in arw's head
 *		added radix sensitive output with IDISP;
 */

STRING errflag;
L_INT  dot;
INT dotinc;
L_INT var[];
char *IDISP;


#define IO 2
#define SEG 1
#define MRNORM 3
#define MRIMM 4
#define MRUNOP 5
#define ACIMM 6
#define SHIFT 7
#define WORD 8
#define WORD2 9
#define REGUNOP 10
#define DISP1 11
#define DISP2 12
#define DISP4 13
#define DATA2 14
#define STRNG 15
#define MRIMS 16
#define MRESC 17
#define MOVI  20
#define MRSEG 21
#define INTR  22
#define ACMEM 23
#define MRJUMP 24                              /* M004 */
#define LEA 25

struct optab {
	char maskl, maskh;
	char vall, valh;
	int itype;
	char *iname;
} optab [] = {
0x03, 0xff, 0x88, 0x00, MRNORM, "mov",
0x01, 0xc7, 0xc6, 0x00, MRIMM,  "mov",
0x0f, 0xff, 0xb0, 0x00, MOVI,   "mov",
0x01, 0xff, 0xa0, 0x00, ACMEM,  "mov",
0x01, 0xff, 0xa2, 0x00, ACMEM,  "mov",
0x00, 0xdf, 0x8e, 0x00, MRSEG,  "mov",
0x00, 0xdf, 0x8c, 0x00, MRSEG,  "mov",

0x00, 0xc7, 0xff, 0x30, MRUNOP, "push",
0x07, 0x00, 0x50, 0x00, REGUNOP,"push",
0x18, 0x00, 0x06, 0x00, SEG,    "push",

0x00, 0xc7, 0x8F, 0x00, MRUNOP, "pop",			/* M001 */
0x07, 0x00, 0x58, 0x00, REGUNOP,"pop",
0x18, 0x00, 0x07, 0x00, SEG,    "pop",

0x00, 0x00, 0x90, 0x00, WORD,	"nop",			/* M004 */
0x01, 0xff, 0x86, 0x00, MRNORM, "xchg",
0x07, 0x00, 0x90, 0x00, REGUNOP,"xchg",

0x01, 0xff, 0xe4, 0x00, IO,     "in",
0x01, 0x00, 0xec, 0x00, STRNG, "in",

0x01, 0xff, 0xe6, 0x00, IO,	"out",
0x01, 0x00, 0xee, 0x00, STRNG,	"out",

0x00, 0x00, 0xd7, 0x00, WORD,   "xlat",
0x00, 0xff, 0x8d, 0x00, LEA,	"lea",
0x00, 0xff, 0xc5, 0x00, LEA,	"lds",
0x00, 0xff, 0xc4, 0x00, LEA,	"les",
0x00, 0x00, 0x9f, 0x00, WORD,   "lahf",
0x00, 0x00, 0x9e, 0x00, WORD,   "sahf",
0x00, 0x00, 0x9c, 0x00, WORD,   "pushf",
0x00, 0x00, 0x9d, 0x00, WORD,   "popf",

0x03, 0xff, 0x00, 0x00, MRNORM, "add",
0x03, 0xc7, 0x80, 0x00, MRIMS,  "add",
0x01, 0xff, 0x04, 0x00, ACIMM,  "add",

0x03, 0xff, 0x10, 0x00, MRNORM, "adc",
0x03, 0xc7, 0x80, 0x10, MRIMS,  "adc",
0x01, 0xff, 0x14, 0x00, ACIMM,  "adc",

0x01, 0xc7, 0xfe, 0x00, MRUNOP, "inc",
0x07, 0x00, 0x40, 0x00, REGUNOP,"inc",
0x00, 0x00, 0x37, 0x00, WORD,   "aaa",
0x00, 0x00, 0x27, 0x00, WORD,   "daa",

0x03, 0xff, 0x28, 0x00, MRNORM, "sub",
0x03, 0xc7, 0x80, 0x28, MRIMS,  "sub",
0x01, 0xff, 0x2c, 0x00, ACIMM,  "sub",

0x03, 0xff, 0x18, 0x00, MRNORM, "sbb",
0x03, 0xc7, 0x80, 0x18, MRIMS,  "sbb",
0x01, 0xff, 0x1c, 0x00, ACIMM,  "sbb",

0x01, 0xc7, 0xfe, 0x08, MRUNOP, "dec",
0x07, 0x00, 0x48, 0x00, REGUNOP,"dec",
0x01, 0xc7, 0xf6, 0x18, MRUNOP, "neg",

0x03, 0xff, 0x38, 0x00, MRNORM, "cmp",
0x03, 0xc7, 0x80, 0x38, MRIMS,  "cmp",
0x01, 0xff, 0x3c, 0x00, ACIMM,  "cmp",

0x00, 0x00, 0x3f, 0x00, WORD,   "aas",
0x00, 0x00, 0x2f, 0x00, WORD,   "das",

0x01, 0xc7, 0xf6, 0x20, MRUNOP, "mul",
0x01, 0xc7, 0xf6, 0x28, MRUNOP, "imul",
0x00, 0x00, 0xd4, 0x0a, WORD2,  "aam",
0x01, 0xc7, 0xf6, 0x30, MRUNOP, "div",
0x01, 0xc7, 0xf6, 0x38, MRUNOP, "idiv",
0x00, 0x00, 0xd5, 0x0a, WORD2,  "aad",
0x00, 0x00, 0x98, 0x00, WORD,   "cbw",
0x00, 0x00, 0x99, 0x00, WORD,   "cwd",

0x01, 0xc7, 0xf6, 0x10, MRUNOP, "not",
0x03, 0xc7, 0xd0, 0x20, SHIFT,  "shl",
0x03, 0xc7, 0xd0, 0x28, SHIFT,  "shr",
0x03, 0xc7, 0xd0, 0x38, SHIFT,  "sar",
0x03, 0xc7, 0xd0, 0x00, SHIFT,  "rol",
0x03, 0xc7, 0xd0, 0x08, SHIFT,  "ror",
0x03, 0xc7, 0xd0, 0x10, SHIFT,  "rcl",
0x03, 0xc7, 0xd0, 0x18, SHIFT,  "rcr",

0x03, 0xff, 0x20, 0x00, MRNORM, "and",
0x01, 0xc7, 0x80, 0x20, MRIMM,  "and",
0x01, 0xff, 0x24, 0x00, ACIMM,  "and",

0x01, 0xff, 0x84, 0x00, MRNORM, "test",
0x01, 0xc7, 0xf6, 0x00, MRIMM,  "test",
0x01, 0xff, 0xa8, 0x00, ACIMM,  "test",

0x03, 0xff, 0x08, 0x00, MRNORM, "or",
0x01, 0xc7, 0x80, 0x08, MRIMM,  "or",
0x01, 0xff, 0x0c, 0x00, ACIMM,  "or",

0x03, 0xff, 0x30, 0x00, MRNORM, "xor",
0x01, 0xc7, 0x80, 0x30, MRIMM,  "xor",
0x01, 0xff, 0x34, 0x00, ACIMM,  "xor",

0x00, 0x00, 0xf3, 0x00, WORD,   "repnz",
0x00, 0x00, 0xf2, 0x00, WORD,   "rep",
0x01, 0x00, 0xa4, 0x00, STRNG, "movs",
0x01, 0x00, 0xa6, 0x00, STRNG, "cmps",
0x01, 0x00, 0xae, 0x00, STRNG, "scas",
0x01, 0x00, 0xac, 0x00, STRNG, "lods",
0x01, 0x00, 0xaa, 0x00, STRNG, "stos",

0x00, 0xff, 0xe8, 0x00, DISP2,  "call",
0x00, 0xc7, 0xff, 0x10, MRJUMP, "call",			/* M004 */
0x00, 0xff, 0x9a, 0x00, DISP4,  "calli",		/* M004 */
0x00, 0xc7, 0xff, 0x18, MRJUMP, "calli",		/* M004 */

0x00, 0xff, 0xe9, 0x00, DISP2,  "jmp",
0x00, 0xff, 0xeb, 0x00, DISP1,  "jmp",
0x00, 0xc7, 0xff, 0x20, MRJUMP, "jmp",			/* M004 */
0x00, 0xff, 0xea, 0x00, DISP4,  "jmpi",			/* M004 */
0x00, 0xc7, 0xff, 0x28, MRJUMP, "jmpi",			/* M001 *//* M004 */

0x00, 0x00, 0xc3, 0x00, WORD,   "ret",
0x00, 0xff, 0xc2, 0x00, DATA2,  "ret",
0x00, 0x00, 0xcb, 0x00, WORD,   "reti",			/* M004 */
0x00, 0xff, 0xca, 0x00, DATA2,  "reti",			/* M004 */

0x00, 0xff, 0x74, 0x00, DISP1,  "je",
0x00, 0xff, 0x7c, 0x00, DISP1,  "jl",
0x00, 0xff, 0x7e, 0x00, DISP1,  "jle",
0x00, 0xff, 0x72, 0x00, DISP1,  "jb",
0x00, 0xff, 0x76, 0x00, DISP1,  "jbe",
0x00, 0xff, 0x7a, 0x00, DISP1,  "jp",
0x00, 0xff, 0x70, 0x00, DISP1,  "jo",
0x00, 0xff, 0x78, 0x00, DISP1,  "js",
0x00, 0xff, 0x75, 0x00, DISP1,  "jne",
0x00, 0xff, 0x7d, 0x00, DISP1,  "jge",
0x00, 0xff, 0x7f, 0x00, DISP1,  "jg",
0x00, 0xff, 0x73, 0x00, DISP1,  "jae",
0x00, 0xff, 0x77, 0x00, DISP1,  "ja",
0x00, 0xff, 0x7b, 0x00, DISP1,  "jnp",
0x00, 0xff, 0x71, 0x00, DISP1,  "jno",
0x00, 0xff, 0x79, 0x00, DISP1,  "jns",
0x00, 0xff, 0xe2, 0x00, DISP1,  "loop",
0x00, 0xff, 0xe1, 0x00, DISP1,  "loope",
0x00, 0xff, 0xe0, 0x00, DISP1,  "loopne",
0x00, 0xff, 0xe3, 0x00, DISP1,  "jcxz",

0x00, 0xff, 0xcd, 0x00, INTR,	"int",
0x00, 0x00, 0xcc, 0x00, WORD,   "int",
0x00, 0x00, 0xce, 0x00, WORD,   "into",
0x00, 0x00, 0xcf, 0x00, WORD,   "iret",

0x00, 0x00, 0xf8, 0x00, WORD,   "clc",
0x00, 0x00, 0xf5, 0x00, WORD,   "cmc",
0x00, 0x00, 0xf9, 0x00, WORD,   "stc",
0x00, 0x00, 0xfc, 0x00, WORD,   "cld",
0x00, 0x00, 0xfd, 0x00, WORD,   "std",
0x00, 0x00, 0xfa, 0x00, WORD,   "cli",
0x00, 0x00, 0xfb, 0x00, WORD,   "sti",
0x00, 0x00, 0xf4, 0x00, WORD,   "hlt",
0x00, 0x00, 0x9b, 0x00, WORD,   "wait",
0x07, 0xc7, 0xd8, 0x00, MRESC,  "esc",
0x00, 0x00, 0xf0, 0x00, WORD,   "lock",
0x18, 0x00, 0x26, 0x00, SEG,    "seg",



0,0,0,0,0,0
};

char *wregs[] = {"ax", "cx", "dx", "bx", "sp", "bp", "si", "di"} ;
char *bregs[] = {"al", "cl", "dl", "bl", "ah", "ch", "dh", "bh"} ;
char *sregs[] = {"es", "cs", "ss", "ds"};

POS space,incp;

printins(f,idsp,ins)
{
    incp = 1;
    space = idsp;
    crack (ins&255);
    dotinc = incp;
}


readb()
{
	return(chkgetb(inkdot(incp++),space));
}

readb2()
{
	int w;
	w = chkget(inkdot(incp),space);
	incp += 2;
	return (w);
}


/*
 * basicly, the scheme is to look at the first byte.  We look through
 * the instuction table until a match is found, masking out potential
 * operands.  if a match is found, and it is of a two byte instruction,
 * we read the next byte and try to match it too.  This algorithim is
 * slow, as it is linear search, but as it will be used user interactive,
 * it is not really that bad.
 * We could speed it up by expanding the table to contain entrys for all
 * first bytes, and then doing the sort of thing we are doing now for
 * the second byte matching, but that is more complicated and takes both
 * bigger code and table.
 */

crack(ins)
{
    struct optab *p;
    int got2;
    POS  l;
    int i,mask;
    int	ins2;

    got2 = 0;
    mask = 0;
    for (p=optab;p->iname!=0;p++)
    {
	if (((~p->maskl & ins)&255) == (p->vall&255))
	{
	    switch (p->itype)
	    {
	case WORD:
		name (p,1);
		return;
	case DISP1:
		name (p,1);
		if (space == NSP)
		{
		    printf ("?");
		    return;
		}
		i = readb();
		if (i&0200) i |= 0177400;
		i += shorten (dot) + shorten(incp);
		psymoff (leng(i), ISYM, "");
		return;
	case DISP2:
		name (p,1);
		if (space == NSP)
		{
		    printf ("?");
		    return;
		}
		i = readb2();
		i += shorten (dot) + shorten (incp);
		psymoff (leng(i), ISYM, "");
		return;
	case DISP4:
		name (p,1);
		i = readb2();				/* M005 */
		hex (l=readb2(), 0);			/* M003 M005*/
		printf (",");
		hex (i, 0);				/* M003 M005*/
		printf(": ");
#ifdef SEGSYM
		psymoff((((long)l)<<16)+i, ISYM, "");	/* M007*/
#else
		psymoff((((long)l)<<4)+i, ISYM, "");	/* M005 M007*/
#endif
		return;
	case DATA2:
		name (p, 1);
		hex (readb2(), 0);			/* M003 */
		return;
	case STRNG:
#ifdef NEVER						/* M008 */
		i = p->iname[3];			/* M002 */
		if ((ins&01) == 0)			/* M002 */
			p->iname[3] = '\0';		/* M002 */
#endif
		name (p, ins);
/*		p->iname[3] = i;			/* M002 M008*/
		return;

	case IO:
		name (p, ins);
		hex (readb(), 0);			/* M003 */
		return;

	case INTR:
		i = readb();
		if ((i&0xf0) != 0xf0) {
			/* isnt floating point trap */
			name (p, 1);
			hex (i, 0);                     /* M003 */
			return;
		}
		printf ("E");
		ins = i;
	case MRESC:
		fpdecode(ins);
		return;

	case SEG:
		name (p, 1);
		printf (sregs[(ins>>3)&3]);
		return;

	case REGUNOP:
		name (p, 1);
		/* xchx is the one byte unop with two ops! -Hans */
		if (p->iname[0] == ' ') printf("ax,");
		printf (wregs[ins&7]);
		return;

	case MRNORM:
		if (!got2) ins2 = readb();
		got2 = 1;
		name (p, ins);
		/* pick direction.		/*M006*/
		if ((ins>>1)&1){
		    preg(ins, ins2);
		    printf (",");
		    indirect (ins, ins2);
		    return;
		}
		indirect (ins, ins2);
		printf (",");
		preg(ins, ins2);
		return;

	case LEA: /*M006*/
		if (!got2) ins2 = readb();
		got2 = 1;
		name (p, ins|1);
	        preg(ins|1, ins2);
	        printf (",");
	        indirect (ins, ins2);
	        return;

	case MOVI:
		name (p, ins>>3);
		if (ins&0x8) printf (wregs[ins&7]);
		else printf (bregs [ins&7]);
		printf (",");
		if (ins&0x8)
		    hex (readb2(), 2);			/* M003 */
		else
		    hex (readb(), 1);			/* M003 */
		return;

	case ACIMM:
		name (p, ins);
		preg (ins, 0);				/* M000 */
		printf (",");				/* M000 */
		if (ins&1)
		    hex (readb2(), 2);			/* M003 */
		else
		    hex (readb(), 1);			/* M003 */
		return;

	case ACMEM:
		name (p, ins);
		/* begin M000 ... */
		if (~ins&2) {
		    preg (ins, 0);
		    printf(",");
		}
	        psymoff (leng(readb2()), DSYM, "");
		if (ins&2) {
		    printf (",");
		    preg(ins, 0);
		}					/* ... end M000 */
		return;

	default:
		if (got2==0) ins2 = readb();
		got2 = 1;
		if (((~p->maskh & ins2)&255) == (p->valh&255))
		{
		    switch (p->itype)
		    {
	case MRIMS:
			mask=2;
	case MRIMM:
			mask++;
			name (p, ins);
			indirect (ins, ins2);
			printf (",");
			if ((ins&mask) == 01)
			    hex (readb2(), 2);			/* M003 */
			else
			    hex (readb(), 1);			/* M003 */
			return;
	case MRUNOP:
	case MRJUMP:						/* M004 */
			name (p, (p->itype == MRUNOP)?ins:1);
			if (p->itype == MRJUMP)			/* M004 */
				printf("@");			/* M004 */
			indirect (ins, ins2);
			return;
	case MRSEG:
			name (p, 1);
			if (ins&2)
			{
			    printf (sregs[(ins2>>3)&3]);
			    printf (",");
			    indirect (ins | 1, ins2);		/* M004 */
			    return;
			}
			indirect (ins | 1, ins2);		/* M004 */
			printf (",");
			printf (sregs[(ins2>>3)&3]);
			return;

	case SHIFT:
			name (p, ins);
			indirect (ins, ins2);
			if (ins&2)
			    printf(",cl");
			else printf (",*1");
			return;

	case WORD2:
			name (p, 1);
			return;

	default:
			printf ("%x %x\t%s\n", ins, ins2, p->iname);
			return;
		    }
		}
	    }
	}
    }
    if (got2==0) ins2 = readb();
    printf ("%x %x ???\n", ins, ins2);
    incp = 1;
}

char *fpstr1[8] = {
	"fadd",
	"fmul",
	"fcom",
	"fcomp",
	"fsub",
	"fsubr",
	"fdiv",
	"fdivr",
	};

char *fpstr2[8] = {
	"fld",
	"ILLEGAL",
	"fst",
	"fstp",
	"ILLEGAL",
	"fldcw",
	"ILLEGAL",
	"fstdw"
	};

char *fpstr3[8] = {
	"fild",
	"ILLEGAL",
	"ILLEGAL",
	"fistp",
	"ILLEGAL",
	"fld",
	"ILLEGAL",
	"fstp"
	};

char *fpstr4[8] = {
	"f16ld",
	"ILLEGAL",
	"ILLEGAL",
	"f16stp",
	"ILLEGAL",
	"ILLEGAL",
	"ILLEGAL",
	"ILLEGAL",
	};



/***    fpdecode - print floating point instruction
 *
 *      fpdecode(ins)
 *
 *      ins = escape code
 */

fpdecode(ins)
    short ins;
{
	register short ins2;
	register short mod;
	register short opc;

	if (ins == 0xf9) {
		printf("fwait");
		return;
	}
	ins2 = readb();
	mod = (ins2>>6) & 3;
	opc = (ins2>>3) & 7;

	switch (ins&7) {

	case 0:                 /* SP form */
	case 4:                 /* DP form */
		printf("%s%c\t", fpstr1[opc], (ins&4)?'d':'s');
		if (mod == 3) {
			if ((ins2 & 0x30) != 0x10)
				printf("ST,");
			printf("ST(%d)", ins2 & 7);
		} else
			indirect(0, ins2);
		return;

	case 1:
		if (mod == 3)
			switch (ins2) {
			case 0xc0:
				printf("fld\tST(0)");
				return;
			case 0xe0:
				printf("fchs");
				return;
			case 0xe1:
				printf("fabs");
				return;
			case 0xfc:
				printf("frndint");
				return;
			default:
				printf("UNKNOWN (%x)", (ins<<8)+ ins2);
				return;
			}
		printf ("%ss\t", fpstr2[opc]);
		indirect(0, ins2);
		return;

	case 2:
		printf ("ILLEGAL\t");
		indirect(0, ins2);
		return;

	case 3:
		if (mod == 3) {
			printf("ILLEGAL");
			return;
		}
		printf ("%s\t", fpstr3[opc]);
		indirect(0, ins2);
		return;

	case 5:
		if (mod == 3) {
			printf ("ILLEGAL");
			return;
		}
		printf ("%sd\t", fpstr2[opc]);
		indirect(0, ins2);
		return;

	case 6:
		if (mod != 3) {
			printf("ILLEGAL");
			indirect(0, ins2);
			return;
		}
		if ((opc == 2) || (opc == 3)) {
			printf("ILLEGAL");
			return;
		}
		printf ("%s\tST(%d),ST", fpstr1[opc], ins2&7);
		return;

	case 7:
		if (mod == 3) {
			printf("ILLEGAL");
			return;
		}
		printf ("%s\t", fpstr4[opc]);
		indirect(0, ins2);
		return;
	}
}



preg(w,i)
char w,i;
{
    if (w&1) printf (wregs[(i>>3)&7]);
    else printf (bregs[(i>>3)&7]);
}

char *mode[] ={
	"(bx)+(si)",
	"(bx)+(di)",
	"(bp)+(si)",
	"(bp)+(di)",
	"(si)",
	"(di)",
	"(bp)",
	"(bx)"
};


indirect (w, i)
{
    int mod, rm, j;

    mod = (i>>6)&3;
    rm = i&7;
    switch (mod)
    {
case 0:
	if (rm==6)
	    psymoff (leng(var[1]=readb2()), DSYM, "");
	else
	    printf (mode[rm]);
	break;
case 1:
	j = readb();
	if (j&0200) j |= 0177400;
	printf ("*%0d.", j);
	printf (mode[rm]);
	break;
case 2:
	printf ("#%0d.", readb2());
	printf (mode[rm]);
	break;
case 3:

	if (w&1) printf (wregs[rm]);
	else printf (bregs[rm]);
    }
}

name (p, b)
struct optab *p;
{
    printf (p->iname);
    if (~b&1) printf ("b");
    printf ("\t");
}

static	char	*hexchar = "0123456789abcdef";	/* begin M003 ... */

/*
 * hex prints a hex value for the disassembler.
 * t indicates the type of value this is
 *	0 - displacement
 *	1 - short immediate
 *	2 - immediate
 */
hex(c, t)
register c, t;
{
	static char s[5];
	if (t)
		printf("%s", t==1? "*": "#");
	t = 4;
	printf(IDISP, c);		/*M008*/
	/*
	do {
		s[--t] = hexchar[ c & 0xF ];
		c >>= 4;
	} while (t && c);
	
	printf ("/%s", &s[t]);
	*/
}					/* ... end M003 */
