/*
 *
 *	XENIX debugger
 *
 *	arw	 7 May 1982	M000
 *		- Changed the name of `pc' to `ip'.  `pc' is now
 *		  just used internally.
 *	arw	 6 Aug 1982	M001
 *		- Fixed printfregs to display 8087 status.
 *	arw	24 Jan 1983	M002
 *		- Added printout of u_fper to printfregs.
 *	Hans	5 Feb, 1983	M003
 *		- middle model $c handling
 *		- print out callers address
 *		- translation into C from SRB language
 *	Hans	2 Mar, 1983	M004
 *		- changed args of valpr (findsym)
 *	Hans	9 Mar 1983	M005
 *		- Added a parameter for size of reglist
 *		- fixed a bug in the search for a register name
 *		- changed the register print order for $r
 *		- Added IDISP and hex IO formats
 */

#include "defs.h"


MSG		LONGFIL;
MSG		NOTOPEN;
MSG		A68BAD;
MSG		A68LNK;
MSG		BADMOD;

MAP		txtmap;
MAP		datmap;

SYMTAB		symbol;
INT		lastframe;
INT		callpc;

INT		infile;
INT		outfile;
CHAR		*lp;
INT		maxoff;
INT		maxpos;
INT		octal;
INT		hex;			/*M005*/
CHAR		*IDISP = "%d.";		/*M005*/

/* symbol management */
L_INT		localval;

/* breakpoints */
BKPTR		bkpthead;

/*begin M005*/
#define REGLMAX 14
REGLIST reglist [] = {
		"ax", &ax,	"fl", &flag,
		"bx", &bx,	"ip", &ip,	/* M000 */
		"cx", &cx,	"cs", &cs,
		"dx", &dx,	"ds", &ds,
		"si", &si,	"ss", &ss,
		"di", &di,	"es", &es,
		"bp", &bp,	"sp", &sp,
};
/*end M005*/


char		lastc;
POS		corhdr[];
POS             *endhdr; /* endhdr used to be a copy of the top of the user
			    stack.  The registers are now kept in vars of
			    their own */

INT		fcor;
STRING		errflg;
INT		signo;


L_INT		dot;
L_INT		var[];
STRING		symfil;
STRING		corfil;
INT		pid;
L_INT		adrval;
INT		adrflg;
L_INT		cntval;
INT		cntflg;

STRING          signals[] = {
		"",
		"hangup",
		"interrupt",
		"quit",
		"illegal instruction",
		"trace/BPT",
		"IOT",
		"EMT",
		"floating exception",
		"killed",
		"bus error",
		"memory fault",
		"bad system call",
		"broken pipe",
		"alarm call",
		"terminated",
};




/* general printing routines ($) */

printtrace(modif)
{
	INT		narg, i, stat, name, limit;
	POS		dynam;
	REG BKPTR	bkptr;
	CHAR		hi, lo;
	INT		word;
	STRING		comptr;
	L_INT		argp, frame, link;
	SYMPTR		symp;

	if(cntflg==0)
		cntval = -1;

	switch (modif) {

	    case '<':
	    case '>':
		{CHAR		file[64];
		INT		index;

		index=0;
		if(modif=='<')
			iclose();
		else
			oclose();
		if(rdc()!=EOR)
		{
			do {
				file[index++]=lastc;
				if(index>=63)
					error(LONGFIL);
			}
			while(readchar()!=EOR);

			file[index]=0;
			if(modif=='<')
			{
				infile=open(file,0);
				if(infile<0)
				{
					infile=0;
					error(NOTOPEN);
				}
			}
			else
			{
				outfile=open(file,1);
				if(outfile<0)
					outfile=creat(file,0644);
				else
					lseek(outfile,0L,2);
			}

		}
		lp--;
		}
		break;

	    case 'o':
		octal = TRUE;
		hex = FALSE;		/*M005*/
		IDISP="%o";
		break;

	    case 'd':
		octal = FALSE;
		hex = FALSE;
		IDISP="%d";		/*M005*/
		break;

	    case 'x':			/*M005*/
		octal = FALSE;
		hex = TRUE;
		IDISP="%x";		/*M005*/
		break;

	    case 'q': case 'Q': case '%':
		done();

	    case 'w': case 'W':
		maxpos=(adrflg?adrval:MAXPOS);
		break;

	    case 's': case 'S':
		maxoff=(adrflg?adrval:MAXOFF);
		break;

	    case 'v': case 'V':
		prints("variables\n");
		for(i=0;i<=35;i++)
			if(var[i])
			{
				printc((i<=9 ? '0' : 'a'-10) + i);
				printf(" = %Q\n",var[i]);
		   	}
		break;

	    case 'm': case 'M':
		printmap("? map",&txtmap);
		printmap("/ map",&datmap);
		break;

	    case 0: case '?':
		if(pid)
			printf("pcs id = %d\n",pid);
		else
			prints("no process\n");
		sigprint();
		flushbuf();

	    case 'r': case 'R':
		printregs();
		return;

	    case 'f': case 'F':
		printfregs(modif=='F');
		return;

	    case 'c': case 'C':
		frame=(adrflg?adrval:bp)&EVEN; lastframe=0;
		callpc=(adrflg?get(frame+2,DSP):pc);
		while(cntval--)
		{
			chkerr();
			narg = findroutine(frame);
			printf("%.8s(", symbol.symc);
			argp = frame+4;
			/* if medium model, findroutine returns
			   negative arg count */
			if (narg<0)
			{			/*M003*/
				narg= -narg;
				argp+=2;
			}
			if(--narg >= 0)
				printf(IDISP, get(argp, DSP));
			while(--narg >= 0)
			{
				argp += 2;
				printf(", ");
				printf(IDISP, get(argp, DSP));
			}
			prints(")\tfrom ");	/*M003*/
			psymoff(leng(callpc-5), ISYM, "");
			prints("\n");

#ifdef LOCALVARS
			/* local variables.  not supported by the
			    compiler, so why bother? M003 */
			if(modif=='C')
			    while( localsym(frame)
			    {
				word=get(localval,DSP);
				printf("%8t%.8s:%10t", symbol.symc);
				if(errflg)
				{
					prints("?\n");
					errflg=0;
				}
				else
					printf("%o\n",word);
			     }
#endif

			lastframe=frame;
			frame=get(frame, DSP)&EVEN;
			if(frame==0) break;
		}
		break;

	    /*print externals*/
	    case 'e': case 'E':
		symset();
		while(symp=symget(0))
		{
			chkerr();
			if((symp->symf)==043 || (symp->symf)==044)
				printf("%.8s:%12t%o\n",
					 symp->symc, get(leng(symp->symv),DSP));
		}
		break;

	    /*set default c frame*/
	    /*print breakpoints*/
	    case 'b': case 'B':
		printf("breakpoints\ncount%8tbkpt%24tcommand\n");
		for(bkptr=bkpthead; bkptr; bkptr=bkptr->nxtbkpt)
		    if(bkptr->flag)
		    {
			printf("%-8.8d",bkptr->count);
			psymoff(leng(bkptr->loc),ISYM,"%24t");
			comptr=bkptr->comm;
			while(*comptr)
				printc(*comptr++);
		    }
		break;

	    default: error(BADMOD);
	}

}

printmap(s,amap)
STRING	s; MAP *amap;
{
	int file;
	file=amap->ufd;
	printf("%s%12t`%s'\n",s,(file<0 ? "-" : (file==fcor ? corfil : symfil)));
	printf("b1 = %-16Q",amap->b1);
	printf("e1 = %-16Q",amap->e1);
	printf("f1 = %-16Q",amap->f1);
	printf("\nb2 = %-16Q",amap->b2);
	printf("e2 = %-16Q",amap->e2);
	printf("f2 = %-16Q",amap->f2);
	printc(EOR);
}

/* Begin M001 ...  */
printfregs(longpr)
{
#ifdef HAVEFP
	REG i, *j;
	INT tagshft, tag;
	L_REAL f;
	static STRING tagname[] = {
		"Valid",
		"Zero",
		"NAN",
		"Empty",
	};

	if(pid)
		readfregs();
	printf("fper = %x\t", corhdr[0].u_fper);
	printf("fpsr = %x\tfpctl = %x\n", i = corhdr[0].u_fps.fp_stat,
		corhdr[0].u_fps.fp_cntrl);
	j = (int *) (corhdr[0].u_fps.fp_regs);
	tagshft = (i>>10) & 0xE;
	tag = corhdr[0].u_fps.fp_tag;

	for(i=0; i<FRMAX; i++)
	{
		printf("st(%d)\t%x\t%x\t%x\t%x\t%x",i,j[0],j[1],j[2],j[3],j[4]);
		printf("\t%s\n", tagname[ (tag>>tagshft) & 03]);
		j += 5;
		tagshft += 2;
		if(tagshft >= 16)
			tagshft = 0;
	}
#endif
}
/*	... end M001	*/

printregs()
{
	REG REGPTR	p;
	INT		v;
	INT		i;

	for(p=reglist,i=0; p < &reglist[REGLMAX]; p++,i++)
	{
		printf("%s%8t", p->rname);		/*M005*/
		printf(IDISP, v= *(p->roffs));
		printf("%8t");
		valpr((long)v,(p->roffs== &ip?ISYM:DSYM));	/* M000,M004*/
		/* put two registers on a line */
		if (i&1) printc(EOR);
		else printf("\t");
	}
	printpc();
}

unsigned int *
getreg(regnam)
{
	REG REGPTR	p;
	REG STRING	regptr;
	CHAR		regnxt;
	regnxt=readchar();
	for(p=reglist; p<&reglist[REGLMAX]; p++)
	{
		regptr=p->rname;
			/* changed || to && M005*/
		if((regnam == *regptr++) && (regnxt == *regptr))
			return(p->roffs);
	}
	lp--;
	return(0);
}

printpc()
{
	dot=pc;
	psymoff(dot,ISYM,":%16t");
	printins(0,ISP,chkget(dot,ISP));
	printc(EOR);
}

sigprint()
{
	prints(signals[signo]);
}
